#!/bin/bash

# Copyright (C) 2023 - 2024 V. Salvatore <vs@salvatorenco.fr>
# Licensend under the GPLv3

VERSION=0.5.1

CFG=${CFG:=/etc/pkg/pkg.conf}

help() {
	echo "Usage: $(basename $0) <function> [package]..."
	echo "$(compgen -A function)" | sed -e 's/^/\t/g' -e 's/$/ .../g'
	exit 1
}

version() {
	echo "pkg $VERSION"
}

# Avoid interpretting untrusted user-input with 'echo -n'
echo_debug() {
	echo -ne "\033[1;36m[DEBUG] \033[0m"
	echo "$@"
}

echo_error() {
	echo -ne "\033[1;31m[ERROR] \033[0m"
	echo "$@"
}

echo_info() {
	echo -ne "\033[1;32m[INFO] \033[0m"
	echo "$@"
}

echo_warning() {
	echo -ne "\033[1;33m[WARNING] \033[0m"
	echo "$@"
}

not_implemented() {
	echo_error "FIXME not implemented"
}

# Helper to sign packages easily. Asks you to choose a private
# key to use when there are multiple on the system.
sign() {
	local pkglist="$(echo $@ | sort | uniq)"

	[ "$#" -eq 0 ] && echo_error "sign requires paths to files to sign" && exit 1

	absent_files="$(ls -- $pkglist 2>&1 | grep 'cannot access' | cut -d\' -f2 | tr '\n' ' ')"
	[ -n "$absent_files" ] && \
			echo_error "Cannot sign absent files '$absent_files'. Aborting" && \
			exit 1

	if [ -z "$MAINTAINER_KEY" ]; then
		private_keys=$(gpg --list-secret-keys | sed -n 's/.*<\(.*\)>$/\1/p')
		PS3="Choose a pgp key to sign: "
		select MAINTAINER_KEY in $private_keys; do
			echo_info "Using $MAINTAINER_KEY key to sign packages";
			break;
		done
	fi

	for x in $pkglist; do
		echo_info "Signing package $x"
		gpg -q --default-key $MAINTAINER_KEY --detach-sign $x
	done
}

download_package_index() {
	[ ! -f "$CACHEDBPATH" ] && mkdir -p $CACHEDBPATH
	[ "$?" != "0" ] && echo_error 'Failed to create cache directory' && exit 1

	echo_info "Downloading package index for your architecture"
	curl --silent -LO --fail ${MIRROR}/binary/${ARCH}/pkgindex --output-dir $CACHEDBPATH/
	[ $? != 0 ] && echo_error "Failed to download pkgindex" && exit 1
}

download() {
	# WARINING: we assume only package names were given, not pkg version
	local packages="$@"
	local arch_combo="${ARCH}/${CPUOPT}/${LIBC}"

	[ -z "$packages" ] && echo_error '<download>: no package name supplied' && exit 1

	if [ -z "$CACHEDBPATH" ]; then
		echo_error '$CACHEDBPATH is not defined. Please specify one.' && exit 1
	elif [ ! -f "$CACHEDBPATH/pkgindex" ]; then
		download_package_index
	fi

	for x in $packages; do

		local the_package=$(grep -- "$arch_combo/${x}-" $CACHEDBPATH/pkgindex | sed "s,.*\(${x}-.*.tar.zst\),\1,g")
		[ -z "$the_package" ] && echo_error "${x} No such package on mirror $MIRROR" && exit 1

		local url=${MIRROR}/binary/$arch_combo/$the_package
		local local_file_size=$([[ -f "$CACHEDBPATH/$the_package" ]] && wc -c < "$CACHEDBPATH/$the_package" || echo "0")
		local remote_file_size=$(curl -sI ${url} | grep 'Content-Length' | cut -d' ' -f2 | tr -d '\r' )

		if [[ "$local_file_size" -ne "$remote_file_size" ]]; then
			echo_info "Downloading ${x}"
			curl -#LO --fail ${url} --output-dir $CACHEDBPATH/ || \
				echo_error "Failed to download package: $x"
				exit 1
		fi
	done
}

install() {
	local rootfs="${flags[root_target]}"

	for pkg in $@; do
		download $pkg

		local the_package=$(grep -- "${ARCH}/${CPUOPT}/${LIBC}/${pkg}-" $CACHEDBPATH/pkgindex | sed "s,.*\(${pkg}-.*.tar.zst\),\1,g")
		local ver="${the_package%.tar*}"; ver=${ver/*-}

		[ $? != 0 ] && echo_error "${pkg} package not found. Did it download successfully?" && exit 1

		# File exists, check if update is available, otherwise do nothing
		if [ -f "${rootfs}/var/adm/packages/$pkg" ]; then
			local installed_version=$(grep "^Package Name and Version:" "$rootfs/var/adm/packages/$pkg" | cut -d' ' -f6)

			if [ "$ver" == "$installed_version" ]; then
				echo_warning "$pkg is already installed. Skipping"
				continue
			else
				# FIXME Actually check for new versions, otherwise we might downgrade
				# packages because we blindly trust the mirror's pkgindex atm

				# FIXME Do not overwrite files if they've been changed by the user.
				# Example: we don't want to break the system
				# by changing /etc/shadow, /etc/fstab
				echo_info "Updating $pkg from $installed_version to $ver"
				tar -C $rootfs -xf $CACHEDBPATH/$the_package
			fi
		else
			echo_info "Installing $pkg $ver"
			tar -C $rootfs -xf $CACHEDBPATH/$the_package
		fi
	done
}

backup() {
	local rootfs="${flags[root_target]}"
	local packages="$@"

	for pkg in $packages; do
		[ ! -f "${rootfs}var/adm/flists/$pkg" ] && \
			echo_warning "$pkg: no such package" 1>&2 && continue

		local ver=$(head -1 ${rootfs}/var/adm/packages/$pkg | cut -d' ' -f6)
		local compressor="zstd -T0 -19"
		local ext="zst"

		echo_info "Creating a binary package of $pkg-$ver"
		(cd ${rootfs}
			(grep ' var/adm' ${rootfs}var/adm/flists/$pkg;
				grep -v ' var/adm' ${rootfs}var/adm/flists/$pkg) | cut -d' ' -f2 \
				| tar -cf- --no-recursion --files-from=- \
				| $compressor
		) > ./$pkg-$ver.tar.$ext
	done
}

detect_host() {
	not_implemented
}

first_configure() {
	not_implemented
}

## MAIN ##
# Check user actually supplied arguments
[ "$#" -eq 0 ] && echo_error "No argument provided." && help

# Map all the functions we delacred so they are easy to reference by user arguments
declare -A func_map
for x in $(compgen -A function); do
	func_map["$x"]="$x"
done

declare -A flags=( \
	["no_config"]="" \
	["show_help"]="" \
	["show_version"]="" \
	["root_target"]="/" \
)

# Parse options. Note that options may be followed by one colon to indicate 
# they have a required argument
if ! options=$(getopt --unquoted -o hnR:V -l help,root:,no-config,version -- "$@")
then
    # Error, getopt will put out a message for us
    exit 1
fi

set -- $options

while [ $# -gt 0 ]
do
	# Consume next (1st) argument
	case $1 in
		-h|--help)
			flags["show_help"]="y" ;;
		-n|--no-config)
			flags["no_config"]="y" ;;
		-V|--version)
			flags["show_version"]="y" ;;

		# Options with required arguments, an additional shift is required
		-R|--root)
			flags["root_target"]="$2"; shift ;;
		--)
			shift; break;;
		-*)
			echo "$0: error - unrecognized option $1" 1>&2; exit 1;;
		*)
			break;;
	esac

	# Fetch next argument as 1st
	shift
done

# We parsed everything, so the remaining $1 should be
# our function name and the rest are package names
MY_CMD=$1
shift

# Do not proceed further if standalone flags (e.g. -h) have been supplied
for flag_name in show_help show_version; do
	case ${flags[$flag_name]} in
		y|yes|1)
			${func_map[${flag_name/show_/}]}
			exit 0
		;;
	esac
done

# Check if the command asked exists
[ -z "${func_map[$MY_CMD]}" ] && \
	echo_error "$MY_CMD function does not exist!" && \
	help

if [ "${flags[no_config]}" != "y" ] ; then
	if [ -f "$CFG" ]; then
		# Load config file
		source $CFG
	else
		echo_error "Config file not found! Please supply --no-config if you want to continue without config file." && exit 1
	fi
fi

# Assume everything else are package names
${func_map[$MY_CMD]} $@
